	/*
 *  MainWindow.cpp
 *  MainWindow
 *
 *  Created by Stefano Morgani on 25/03/09.
 *
 */
 
#include "MainWindow.h"

extern "C" {

    void *run_udp_function(void *obj) {
      return ((MainWindow *)obj)->run_data_rcv_thread_udp((void *) ((MainWindow *)obj)->ui.txtPort->value());
    }

    void *run_tcp_function(void *obj) {
      return ((MainWindow *)obj)->run_data_rcv_thread_tcp((void *) ((MainWindow *)obj)->ui.txtPort->value());
    }

    void *run_fps_function(void *obj) {
      return ((MainWindow *)obj)->run_fps_thread((void *)0);
    }

}

void* MainWindow::run_fps_thread(void* arg) {
    while(1) {
        sleep(TIMER);
        emit new_fps();
        if(restart_tcp_server) {
            restart_tcp_server = false;
            pthread_create(&data_rcv_thread_tcp, NULL, run_tcp_function, this);
        }
    }
}

void MainWindow::encode_int (int nb, char *data, int pos) {

        data[pos] = nb & 0x000000FF;
        data[pos-1] = (nb & 0x0000FF00) >> 8;
        data[pos-2] = (nb & 0x00FF0000) >> 16;
        data[pos-3] = (nb & 0xFF000000) >> 24;
}

int MainWindow::decode_int (char *data, int pos) {
        int res;
        unsigned char tmp_char;

        res = (int) (unsigned char) data[pos];

        tmp_char = (unsigned char) data[pos-1];
        res += (int) tmp_char << 8;

        tmp_char = (unsigned char) data[pos-2];
        res += (int) tmp_char << 16;

        tmp_char = (unsigned char) data[pos-3];
        res += (int) tmp_char << 24;

        return res;
}

void* MainWindow::run_data_rcv_thread_udp(void* arg) {
    long port = (long)arg;

    //printf("port = %d\n", port);

    //int sd;
    int rc;
    socklen_t cliLen;
    struct sockaddr_in cliAddr, servAddr;
    char *buffer = NULL;
    int image_size = 0;
    bool first_image = true;
    int fragment_id = -1;       // image fragment id
    std::ofstream myFile;
    int num_bytes = 0;

    /* socket creation */
    sd=socket(AF_INET, SOCK_DGRAM, 0);
    if(sd < 0) {
        perror("Cannot create socket");
        exit(1);
    }

    /* bind local server port */
    servAddr.sin_family = AF_INET;
    servAddr.sin_addr.s_addr =  htonl(INADDR_ANY);  //inet_addr("192.168.2.3");
    servAddr.sin_port = htons (port);
    rc = bind (sd, (struct sockaddr *) &servAddr, sizeof(servAddr));
    if(rc < 0) {
        perror("Cannot bind");
        exit(1);
    }

    printf("UDP server running on port %d...\n", port);

    abort_udp_thread = false;

    int img_size = 0;

    /* server infinite loop */
    while (!abort_udp_thread) {

        memset(msg, 0x0, MAX_MSG_LEN);                         // reset buffer

        cliLen = sizeof(cliAddr);

        num_bytes = recvfrom(sd, msg, MAX_MSG_LEN, 0, (struct sockaddr *) &cliAddr, &cliLen);

        //std::cerr << "Camera number of bytes received = " << num_bytes << std::endl;
        //std::cerr << "Camera message received: " <<  msg << std::endl;

        if(num_bytes < 0) {
            if(errno != EINTR) {                                    // interrupted system call
                perror("(Camera) Error in receiving...");
            }
            continue;
        }



        fragment_id = decode_int((char*)msg, 3);

        if(first_image) {                                   // if it's the first image we received, then open the output stream (the id = 0, but there is no image to print out)
            first_image = false;
            myFile.open ("camera.jpeg", std::ios::out | std::ios::binary);
            myFile.write ((char*)&msg[4], num_bytes-4);            // write the fragment data (msg length - header = msg type + fragment id)
            if(!myFile) {
                std::cerr << "error in writing image" << std::endl;
                continue;
            }
        } else {
            if(fragment_id == 0) {                          // if a new image arrives, then print out the old one
                myFile.seekp(0);                            // reset the pointer to the begin of the file
                image = QImage("camera.jpeg");
                if(image.isNull()) {
                    std::cerr << "(Camera) my_image is null" << std::endl;
                } else {
                    image = image.transformed(QMatrix(-1.0, 0.0, 0.0, -1.0, 0.0, 0.0));
                    emit new_camera_image(image);
                    //std::cerr << "(Camera) my_image emitted" << std::endl;
                }
            }
            myFile.write ((char*)&msg[4], num_bytes-4);            // write the fragment data (msg length - header = msg type + fragment id)
            if(!myFile) {
                std::cerr << "error in writing image" << std::endl;
                continue;
            }
        }

        //memcpy(image_buffer, msg, img_size);
        //emit image_received_udp();

    }/* end of server infinite loop */

    cleanup(sd);
    //fprintf(stderr, "cleanup udp\n");
    /*
    ui.radioTCP->setEnabled(true);
    ui.radioUDP->setEnabled(true);
    ui.btnConnect->setEnabled(true);
    ui.txtPort->setEnabled(true);
    */
}

void MainWindow::paint_camera_image(QImage img) {
    if(img.isNull()) {
        std::cerr << "null camera image" << std::endl;
        return;
    }

    QLabel *lblCamera = new QLabel(this);
    cam_pix = QPixmap::fromImage(img);
    if(!cam_pix.isNull()) {
        lblCamera->setPixmap(QPixmap::fromImage(img));
        ui.scrollCamera->setWidget(lblCamera);
    }

    if(!abort_udp_thread) {
        images_ok_udp++;
    } else {
        images_ok_tcp++;
    }

}

void* MainWindow::run_data_rcv_thread_tcp(void* arg) {
    long port = (long) arg;

    //printf("port = %d\n", port);

    //int local_id;       //socket id of the local machine (server)
    //int local_copy_id;  //socket id returned from "accept" function
    int ret_value;

    bool first_image = true;
    int fragment_id = -1;       // image fragment id
    std::ofstream myFile;
    int num_bytes = 0;


    struct sockaddr_in sock_local;  //info about local machine
    struct sockaddr_in sock_remote; //info about remote machine
    unsigned int sock_remote_len;

    /* socket creation */
    local_id = socket(AF_INET, SOCK_STREAM, 0);
    if(local_id < 0) {
        perror("Cannot create socket: ");
        exit(1);
    }

    //Il socket deve essere non bloccante
    //ret_value=fcntl(local_id,F_SETFL,O_BLOCK);
    //if(ret_value < 0) {
    //   perror("fnctl: ");
    //}

        int optval;
    optval=1;

    if(setsockopt(local_id, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval))==-1) {
        perror("setsockopt");
    } else {
        //printf("setsockopt ok...");
    }

    /* bind local server port */
    sock_local.sin_family = AF_INET;
    sock_local.sin_addr.s_addr = htonl(INADDR_ANY); //inet_addr("192.168.2.2");  //htonl(INADDR_ANY);  //inet_addr("192.168.1.54");
    sock_local.sin_port = htons (port);

    ret_value = bind (local_id, (struct sockaddr *) &sock_local, sizeof(sock_local));
    if(ret_value < 0) {
        perror("Cannot bind: ");
        exit(1);
    }

    ret_value = listen(local_id, BACKLOG);    //one pending connection
    if(ret_value < 0) {
        perror("listen: ");
        exit(1);
    }

        socklen_t optlen = sizeof(sock_local);

       // getsockopt(sd, SOL_SOCKET, SO_MAX_MSG_SIZE, (int *)&optval, &optlen);
       // printf("Max msg size = %d\n", optval);
        getsockopt(local_id, SOL_SOCKET, SO_SNDBUF, (int *)&optval, &optlen);
        printf("Sending buffer size = %d\n", optval);
        getsockopt(local_id, SOL_SOCKET, SO_RCVBUF, (int *)&optval, &optlen);
        printf("Receiving buffer size = %d\n", optval);




    printf("TCP server running...\n");

    sock_remote_len = sizeof(sock_local);

    local_copy_id = accept(local_id, (struct sockaddr *) &sock_remote, &sock_remote_len);
    if(local_copy_id < 0) {
        perror("accept: ");
        cleanup(local_id);
        cleanup(local_copy_id);
        exit(1);
    }


    printf("Remote address: %s\n", inet_ntoa(sock_remote.sin_addr));
    printf("Remote port = %d\n", htons(sock_remote.sin_port));

    int bytes_received = 0;

    abort_tcp_thread = false;

    bool is_complete = false;
    bool first_fragment = true;
    int img_size = 0;
    int fragment_size=0;

    /* server infinite loop */
    while (!abort_tcp_thread) {

        /* init buffer */
        memset(msg, 0x0, MAX_MSG_LEN);

        bytes_received = 0;



        //num_bytes = read(local_copy_id, &msg[0], 8); //recv(local_copy_id, &msg[bytes_received], MAX_MSG_LEN-bytes_received, 0);
        recv(local_copy_id, &msg[0], 8, MSG_WAITALL);
        if(num_bytes < 0) {
            if(errno != EINTR) {                                    // interrupted system call
                perror("(Camera) Error in receiving...");
            }
            continue;
        }
        fragment_size = decode_int((char*)msg, 7);
        //std::cerr << "fragment size = " << fragment_size << std::endl;
        //std::cerr << "received bytes = " << num_bytes << std::endl;

        num_bytes=0;
        //while(num_bytes < fragment_size) {
            num_bytes += recv(local_copy_id, &msg[8+num_bytes], fragment_size-num_bytes, MSG_WAITALL); //MAX_MSG_LEN-bytes_received, 0);
            //num_bytes += read(local_copy_id, &msg[8+num_bytes], fragment_size-num_bytes); //recv(local_copy_id, &msg[bytes_received], MAX_MSG_LEN-bytes_received, 0);
        //}
        if(num_bytes < 0) {
            if(errno != EINTR) {                                    // interrupted system call
                perror("(Camera) Error in receiving...");
            }
            continue;
        } else if (num_bytes == 0) {
            abort_tcp_thread = true;
        }

        //std::cerr << "received bytes = " << num_bytes << std::endl;

        fragment_id = decode_int((char*)msg, 3);

        if(first_image) {                                   // if it's the first image we received, then open the output stream (the id = 0, but there is no image to print out)
            first_image = false;
            myFile.open ("camera.jpeg", std::ios::out | std::ios::binary);
            myFile.write ((char*)&msg[8], num_bytes);            // write the fragment data (msg length - header = msg type + fragment id)
            if(!myFile) {
                std::cerr << "error in writing image" << std::endl;
                continue;
            }
        } else {
            if(fragment_id == 0) {                          // if a new image arrives, then print out the old one
                myFile.seekp(0);                            // reset the pointer to the begin of the file
                image = QImage("camera.jpeg");
                if(image.isNull()) {
                    std::cerr << "(Camera) my_image is null" << std::endl;
                } else {
                    image = image.transformed(QMatrix(-1.0, 0.0, 0.0, -1.0, 0.0, 0.0));
                    emit new_camera_image(image);
                    //std::cerr << "(Camera) my_image emitted" << std::endl;
                }
            }
            myFile.write ((char*)&msg[8], num_bytes);            // write the fragment data (msg length - header = msg type + fragment id)
            if(!myFile) {
                std::cerr << "error in writing image" << std::endl;
                continue;
            }
        }

/*
        // receive message
        is_complete = false;
        first_fragment = true;
        //while(bytes_received < 3203) {
        while(!is_complete) {
            if(abort_tcp_thread) {
                break;
            }
            ret_value = read(local_copy_id, &msg[bytes_received], MAX_MSG_LEN-bytes_received); //recv(local_copy_id, &msg[bytes_received], MAX_MSG_LEN-bytes_received, 0);
            if(first_fragment) {
                switch((int)msg[0]) {
                    case 0:
                        img_size = ((int)msg[1]*(int)msg[2])+3;
                        break;
                    case 1:
                        img_size = ((int)msg[1]*(int)msg[2]*2)+3;
                        break;
                }
                first_fragment = false;
            }
            fprintf(stderr, "ret = %d, bytes to read = %d\n", ret_value, MAX_MSG_LEN-bytes_received);
            if(ret_value > 0) {
                bytes_received += ret_value;

                switch((int)msg[0]) {
                    case 0:
                        if(bytes_received == img_size ) {
                            is_complete = true;
                        } else if(bytes_received > img_size) {
                            fprintf(stderr, "more than one image received\n");
                        }
                        break;
                    case 1:
                        if(bytes_received == img_size ) {
                            is_complete = true;
                        } else if(bytes_received > img_size) {
                            fprintf(stderr, "more than one image received\n");
                        }
                        break;
                }

            }

        }

        //ret_value = recv(local_copy_id, &msg[0], MAX_MSG_LEN, MSG_WAITALL);   //MSG_WAITALL

        //printf("Bytes received = %d\n", bytes_received);
        //printf("message received: %s\n", msg);

        //if(ret_value < 0) {

         //   perror("\nError in receiving data: ");
         //   continue;

        //} else if (ret_value > 0) {


        //if(ret_value == MAX_MSG_LEN) {
            memcpy(image_buffer, msg, img_size);
            emit image_received_tcp();
        //}

        //}
*/

    }/* end of server infinite loop */

    cleanup(local_id);
    //fprintf(stderr, "cleanup tcp\n");
    /*
    ui.radioTCP->setEnabled(true);
    ui.radioUDP->setEnabled(true);
    ui.btnConnect->setEnabled(true);
    ui.txtPort->setEnabled(true);
    */
    restart_tcp_server = true;
}

void MainWindow::cleanup(int sock) {
    int ret_value;

    ret_value = ::close(sock);

    if(ret_value == -1) {
        perror("\nclose: ");
    }
}

MainWindow::MainWindow(QMainWindow *parent) : QMainWindow(parent) {
	ui.setupUi(this);
        images_lost_udp = 0;
        tot_images_udp = 0;
        images_ok_udp = 0;
        images_lost_tcp = 0;
        tot_images_tcp = 0;
        images_ok_tcp = 0;
        abort_tcp_thread = true;
        abort_udp_thread = true;
        restart_tcp_server = false;
        return;
}

MainWindow::~MainWindow() {
    cleanup(sd);
    cleanup(local_id);
    pthread_cancel(fps_thread);
}

void MainWindow::start_server() {
    if(ui.radioTCP->isChecked()) {
        pthread_create(&data_rcv_thread_tcp, NULL, run_tcp_function, this);
    } else {
        pthread_create(&data_rcv_thread_udp, NULL, run_udp_function, this);
    }
    pthread_create(&fps_thread, NULL, run_fps_function, this);
    ui.radioTCP->setEnabled(false);
    ui.radioUDP->setEnabled(false);
    ui.btnConnect->setEnabled(false);
    ui.txtPort->setEnabled(false);
    //QObject::connect(ui.radioTCP, SIGNAL(clicked()), this, SLOT(init_tcp_socket()));
    //QObject::connect(ui.radioUDP, SIGNAL(clicked()), this, SLOT(init_udp_socket()));
}

void MainWindow::visualize() {

    if(!abort_udp_thread) {
        tot_images_udp++;
    } else {
        tot_images_tcp++;
    }

    fprintf(stderr, "type = %d, width = %d, height = %d\n", image_buffer[0], image_buffer[1], image_buffer[2]);

    if(image_buffer[1] != 0 && image_buffer[2] != 0) {    //print a new image only if received correct data, othrewise leave the old image
        switch(image_buffer[0]) {
            case 0:	{	//grayscale
                image = QImage(image_buffer[1], image_buffer[2], QImage::Format_RGB32);
                int i=3;
                for(int y=0; y<image_buffer[2]; y++) {
                    for(int x=0; x<image_buffer[1]; x++) {
                        int r = (int)image_buffer[i];
                        int g = (int)image_buffer[i];
                        int b = (int)image_buffer[i];
                        image.setPixel(x, y, qRgb (r, g, b));
                        i++;
                    }
                }
                QLabel *lblCamera = new QLabel;
                //the size of the label is 200*200, the maximum size of the image is 45*45
                lblCamera->setFixedSize(image_buffer[1]*200/45,image_buffer[2]*200/45);
                lblCamera->setScaledContents(true);
                lblCamera->setPixmap(QPixmap::fromImage(image));
                ui.scrollCamera->setWidget(lblCamera);
                break;
            }

            case 1:	{	// RGB565 image
                image = QImage(image_buffer[1], image_buffer[2], QImage::Format_RGB16);
                int i=0;
                for(int y=0; y<image_buffer[2]; y++) {
                    for(int x=0; x<image_buffer[1]; x++) {
                        int r = (int)image_buffer[i*2+3]&0xF8;
                        int g = (int)(image_buffer[i*2+3]&0x07)<<5 | (image_buffer[i*2+1+3]&0xE0)>>3;
                        int b = (int)(image_buffer[i*2+1+3]&0x1F)<<3;
                        image.setPixel(x, y, qRgb (r, g, b));
                        i++;
                    }
                }
                QLabel *lblCamera = new QLabel(this);
                lblCamera->setFixedSize(image_buffer[1]*200/45, image_buffer[2]*200/45);
                lblCamera->setScaledContents(true);
                lblCamera->setPixmap(QPixmap::fromImage(image));
                ui.scrollCamera->setWidget(lblCamera);
                break;
            }

            default: break;
        }

        if(!abort_udp_thread) {
            images_ok_udp++;
        } else {
            images_ok_tcp++;
        }

    } else {

        if(!abort_udp_thread) {
            images_lost_udp++;
        } else {
            images_lost_tcp++;
        }

    }

    if(!abort_udp_thread) {
        fprintf(stderr, "UDP total images = %d\n", tot_images_udp);
        fprintf(stderr, "UDP images lost = %d\n", images_lost_udp);
    } else {
        fprintf(stderr, "TCP total images = %d\n", tot_images_tcp);
        fprintf(stderr, "TCP images lost = %d\n", images_lost_tcp);
    }

}

void MainWindow::init_tcp_socket() {
    //pthread_cancel(data_rcv_thread_udp);
    //cleanup(sd);
    abort_udp_thread = true;
    pthread_create(&data_rcv_thread_tcp, NULL, run_tcp_function, this);
    tot_images_tcp = 0;
    images_lost_tcp = 0;
}

void MainWindow::init_udp_socket() {
    //pthread_cancel(data_rcv_thread_tcp);
    //cleanup(local_id);
    abort_tcp_thread = true;
    pthread_create(&data_rcv_thread_udp, NULL, run_udp_function, this);
    tot_images_udp = 0;
    images_lost_udp = 0;
}

void MainWindow::update_fps() {
    if(!abort_udp_thread) {
        fps = (double)images_ok_udp/(double)TIMER;
        images_ok_udp = 0;
    } else {
        fps = (double)images_ok_tcp/(double)TIMER;
        images_ok_tcp = 0;
    }
    std::cerr << "fps = " << fps << std::endl;
    ui.lblFps->setNum(fps);
}
