/*
 *  MainWindow.h
 *  MainWindow
 *
 *  Created by Stefano Morgani on 25/03/09.
 *
 */

/*! \class MainWindow
 *  \brief Main window class
 *  \author Stefano Morgani
 *  \version 1.0
 *  \date 25/03/09
 *
 * description...
 */

#ifndef MAINWINDOW_H_
#define MAINWINDOW_H_

#include "ui_main.h"
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <errno.h>
extern "C" {
	#include <pthread.h>
}
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define MAX_MSG_LEN 8508        ////should be 255*255*2+3, becuase one char (for both width and height) has value at most 255
#define BACKLOG 1
#define TIMER 5

class MainWindow : public QMainWindow
{
	Q_OBJECT
	
	public:
		//functions
		MainWindow(QMainWindow *parent=0);
		~MainWindow();
                void* run_data_rcv_thread_udp(void* arg);       //udp data receiving thread function
                void* run_data_rcv_thread_tcp(void* arg);       //tcp data receiving thread function
                void* run_fps_thread(void* arg);                //thread calculating the frame per second
                void cleanup(int sock);
                void encode_int (int nb, char *data, int pos);
                int decode_int (char *data, int pos);

		//members
                Ui::MainWindow ui;
                pthread_t data_rcv_thread_udp;	//camera receiving thread (receive data from wireless udp connection)
                pthread_t data_rcv_thread_tcp;  //camera receiving thread (receive data from wireless tcp connection)
                pthread_t fps_thread;
                int images_lost_udp;            //images for which the header isn't correct
                int tot_images_udp;             //total number of images received: images_lost_udp + images_ok_udp
                int images_ok_udp;              //images correctly received
                int images_lost_tcp;
                int tot_images_tcp;
                int images_ok_tcp;
                bool abort_tcp_thread;
                bool abort_udp_thread;
                int sd, local_id, local_copy_id;
                unsigned char msg[MAX_MSG_LEN];
                unsigned char image_buffer[MAX_MSG_LEN];
                QImage image;
                QPixmap cam_pix;
                double fps;
                bool restart_tcp_server;

	public slots:
		void start_server();
                void visualize();
                void init_tcp_socket();
                void init_udp_socket();
                void update_fps();
                void paint_camera_image(QImage img);

	signals:
                void image_received_udp();
                void image_received_tcp();
                void new_fps();
                void new_camera_image(QImage img);
};

#endif
